/******************************************************************************
 * compat.cpp - Windows/Linux compatibility grab-bag functions.               *
 *                                                                            *
 * All the mess needed to build my simple programs on both Linux (using gcc)  *
 * and Windows (using MS Visual Studio) is kept here. See also compat.h.      *
 *                                                                            *
 *  Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>               *
 *                                                                            *
 *  compat is free software: you can redistribute it and/or modify it         *
 *  under the terms of the GNU General Public License as published by the     *
 *  Free Software Foundation, either version 3 of the License, or             *
 *  (at your option) any later version.                                       *
 *                                                                            *
 *  compat is distributed in the hope that it will be useful, but             *
 *  WITHOUT ANY WARRANTY; without even the implied warranty of                *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
 *  See the GNU General Public License for more details.                      *
 *                                                                            *
 *  You should have received a copy of the GNU General Public License along   *
 *  with this program.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#include "stdafx.h"

#include <stdio.h>
#include <string.h>
#include <iostream>
#include <fstream>

#include "config.h"
#include "compat.h"


#ifdef UNIX
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include "function.h"

int fopen_s(FILE **f, const char *path, const char *mode) {
   /***************************************************************************
    * This implements Microsoft's new equivalent to fopen so that we can use  *
    * the same call everywhere without warnings in MSVC++.                    *
    *                                                                         *
    * It opens the specified file in the specified mode, sets *f accordingly  *
    * and returns zero. If the operation fails, *f is set NULL and -1 is      *
    * returned.                                                               *
    ***************************************************************************/
   return (*f = fopen(path, mode)) ? 0 : -1; }



void Sleep(int d) {
   /***************************************************************************
    * This implements the Windows Sleep function in Linux. It pauses for d    *
    * milliseconds.                                                           *
    ***************************************************************************/
   usleep(1000 * d); }
void changemode(int dir) {
   /***************************************************************************
    * This should be called with dir=1 before kbhit() to obtain               *
    * non-linebuffered keyboard input in Linux. It should be called again     *
    * afterwards with dir=0 to restore things.                                *
    ***************************************************************************/
   static struct termios oldt, newt;

   if ( dir == 1 ) {
      tcgetattr( STDIN_FILENO, &oldt);
      newt = oldt;
      newt.c_lflag &= ~( ICANON | ECHO );
      tcsetattr( STDIN_FILENO, TCSANOW, &newt); }
   else
      tcsetattr( STDIN_FILENO, TCSANOW, &oldt); }



int _kbhit (void) {
   /***************************************************************************
    * This provides the Windows kbhit() in Linux. It returns TRUE if a key    *
    * has been pressed. Be sure to use changemode(1) before using this and    *
    * changemode(0) once finished.                                            *
    ***************************************************************************/
   struct timeval tv;
   fd_set rdfs;

   tv.tv_sec = 0;
   tv.tv_usec = 0;

   FD_ZERO(&rdfs);
   FD_SET (STDIN_FILENO, &rdfs);

   select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
   return FD_ISSET(STDIN_FILENO, &rdfs); }



#else // nUNIX

 int main(int argc, char *argv[]);                          // defined in dj.cpp

int wmain(int argc, wchar_t *argv[]) {
   /***************************************************************************
    * The Windows entry point is wmain() which provides UTF-16 arguments in   *
    * argv. This veneer converts them to UTF-8 and calls the same main() as   *
    * with Linux. (Note that, if Windows sees main(), it may call it instead  *
    * with ASCII arguments. We avoid this by renaming main() using a #define  *
    * in compat.h.)                                                           *
    *                                                                         *
    * Be sure to define MAX_ARGS (the number of command-line arguments your   *
    * program can take) and MAX_ARG_LEN (the maximum length of any individual *
    * argument) in your config.h. Setting these too high is OK except for a   *
    * (trivial) waste of memory.                                              *
    ***************************************************************************/
   char *argv8p[MAX_ARGS],               // pointers to the converted arguments
      argv8[MAX_ARGS][MAX_ARG_LEN];       // the converted arguments themselves
   int i;                                                            // scratch

   if (argc > MAX_ARGS)
      argc = MAX_ARGS;                                            // for safety

   /*
    * We convert each argument to a UTF-8 string and set a pointer to it in
    * argv8p[].
    */
   for (i = 0; i < argc; i++) {
      WideCharToMultiByte(
         CP_UTF8,                               // UINT CodePage - output UTF-8
         0,                                     //  DWORD dwFlags - none needed
         argv[i],                   // LPCWSTR lpWideCharStr - the input string
         -1,                     //  int cchWideChar - input is null-terminated
         argv8[i],                      // LPSTR lpMultiByteStr - output string
         MAX_ARG_LEN,                 // int cbMultiByte - maximum output chars
         NULL,                               // LPCSTR lpDefaultChar - not used
         NULL );                         // LPBOOL lpUsedDefaultChar - not used
      argv8p[i] = &argv8[i][0]; }

   return main(argc, argv8p); }


void changemode(int dir) {
   /***************************************************************************
    * This function is not needed and does nothing in Windows.                *
    ***************************************************************************/
 }



int ftw(const char *dirpath,
   int (*fn) (const char *fpath, const struct fstat *sb,
      int typeflag),
   int nopenfd) {
   /***************************************************************************
    * In Windows, this partially implements the file tree walk function.      *
    *                                                                         *
    * Starting with the directory at dirpath, it finds subdirectories and     *
    * files. For each file found, it calls fn, passing the full path and file *
    * name in fpath, information about the file (as enabled in config.h, e.g. *
    * ST_SIZE) in sb and typeflag set to FTW_F.                               *
    *                                                                         *
    * When a subdirectory is found, fn is called with its full path in fpath  *
    * and typeflag set to FTW_D, then ftw calls itself recursively to process *
    * the contents of that subdirectory.                                      *
    *                                                                         *
    * If fn ever returns a non-zero value, the walk ends immediately and the  *
    * value from fn is returned. Otherwise, the walk ends and zero is         *
    * returned once all files and subdirectories at the top level (dirpath)   *
    * have been processed.                                                    *
    *                                                                         *
    * -1 is returned if an error occurs such as a path becoming too long.     *
    ***************************************************************************/
   WIN32_FIND_DATA ffd;
   HANDLE dirHandle = INVALID_HANDLE_VALUE;
   char filePath[MAX_FP_CHARS];                // path to the directory or file
//   wchar_t filePathW[MAX_FP_CHARS];                       // filePath as UTF-16

   struct fstat fileStat;           // information about the file that is found
   int
      dirLen,                                              // length of dirpath
      fileFound,           // false once there are no more files or directories
      fnResult = 0,                                           // result from fn
      typeFlag;                                               // FTW_D or FTW_F

   if ((dirLen = strlen(dirpath)) > sizeof filePath - 3)
      return -1;                   // oops - dirpath is too long for the buffer

   /*
    * We set up the buffer, appending "/*" to dirpath. We then convert it
    * from UTF-8 to UTF-16 coding.
    */
   strlcpy(filePath, dirpath, sizeof filePath);
   filePath[dirLen++] = '/';
   filePath[dirLen] = '*';
   filePath[dirLen + 1] = '\0';
//   if (!MultiByteToWideChar(CP_UTF8, 0, filePath, -1, filePathW, MAX_FP_CHARS))
//      return -1;                // conversion failed (path too long?). We quit.

//std::wcout << L"P1 " << filePathW << L"\n";

   /*
    * We find the first subdirectory or file in the directory then, if that
    * succeeds, we process it and subsequent ones.
    */
   dirHandle = FindFirstFile(filePath, &ffd);
   fileFound = dirHandle != INVALID_HANDLE_VALUE;
   while (fileFound && !fnResult) {
   
//std::cout << "P2 " << ffd.cFileName << "\n";
   
      if (strcmp(ffd.cFileName, ".") && strcmp(ffd.cFileName, "..")) {
         /*
          * The subdirectory or file is valid (we ignore . and ..). We
          * convert timestamps as required to time_t form (Windows filetimes
          * are in 100 nsec increments since 1/1/1601 UTC).
          */

#ifdef ST_ATIME
         fileStat.st_atime = (
            ((long long) ffd.ftLastAccessTime.dwHighDateTime << 32)
            + ((long long) ffd.ftLastAccessTime.dwLowDateTime & 0xffffffff)
            - 116444736000000000) / 10000000;
#endif // ST_ATIME

#ifdef ST_CTIME
         fileStat.st_ctime = (
            ((long long) ffd.ftCreationTime.dwHighDateTime << 32)
            + ((long long) ffd.ftCreationTime.dwLowDateTime & 0xffffffff)
            - 116444736000000000) / 10000000;
#endif // ST_CTIME

#ifdef ST_MTIME
         fileStat.st_mtime = (
            ((long long) ffd.ftLastWriteTime.dwHighDateTime << 32)
            + ((long long) ffd.ftLastWriteTime.dwLowDateTime & 0xffffffff)
            - 116444736000000000) / 10000000;
#endif // ST_MTIME

#ifdef ST_HIDDEN
         fileStat.st_hidden =
            (ffd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) != 0;
#endif

         if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            /*
             * We found a subdirectory. We set the type and size accordingly.
             */
             
//std::cout << "P3 is dir\n";
             
            typeFlag = FTW_D;
#ifdef ST_SIZE
            fileStat.st_size = 0;
#endif // ST_SIZE
         }
         else
         {
             
//std::cout << "P4 is file *" << filePath << "* " << sizeof filePath << " *" << ffd.cFileName << "*\n";
             
            /*
             * We found a file. We set the type and put other information
             * about it (currently only its size) in fileStat.
             */
            typeFlag = FTW_F;
#ifdef ST_SIZE
            LARGE_INTEGER fileSize;
            fileSize.LowPart = ffd.nFileSizeLow;
            fileSize.HighPart = ffd.nFileSizeHigh;
            fileStat.st_size = fileSize.QuadPart;
#endif // ST_SIZE
         }

            /*
             * We concatenate the subdirectory or file name with filePath and call fn.
             * We do this for directories as well as files to match the Linux ftw().
             */
         strcpy_s(&filePath[dirLen], sizeof filePath - dirLen, ffd.cFileName);
             
//std::cout << "P5 " << filePath << "\n";
             
            fnResult = fn(filePath, &fileStat, typeFlag);

            if (typeFlag == FTW_D && !fnResult)
               /*
                * That went OK and we have a subdirectory. We call ourselves
                * recursively to dive into it.
                */
               fnResult = ftw(filePath, fn, nopenfd); }

      /*
       * We get the next subdirectory or file.
       */
      fileFound = FindNextFile(dirHandle, &ffd); }

   FindClose(dirHandle);

   return fnResult; }

#endif // nUNIX


int IsHidden(const char *fpath, const struct fstat *sb) {
   /***************************************************************************
    * This returns TRUE if the file or directory is hidden.                   *
    ***************************************************************************/
#ifdef UNIX
   /*
    * A file or directory is hidden if its name starts with '.'. We find the
    * start of any extension.
    */
   int pos = 0;
   PathParse(fpath, 0, &pos);

   if (!PathParse(fpath, 0, &pos))
      /*
       * The name looks to be zero-length because it starts with '.' and
       * PathParse thought it was the extension.
       */
      return 1;

   /*
    * We check the first character of the name.
    */
   return fpath[pos] == '.';

#else
   return sb->st_hidden;                // hidden is a file attribute in Windows
#endif
 }


int FileCopy(const char *d, const char *s) {
   /***************************************************************************
    * This copies the file whose path and name is s as a file whose path and  *
    * name is d. Both s and d are UTF-8 coded. It returns 0 iff successful.   *
    ***************************************************************************/

#ifdef UNIX
   /*
    * We construct an ifstream ss for the file named s. With Linux, it's
    * simple: the ifstream (and ofstream) constructors work as-is with UTF-8
    * path and file names.
    */
   std::ifstream ss(s, std::ios::binary);

#else
   /*
    * The Windows ifstream (and ofstream) constructors expect either ASCII
    * (char*) or UTF-16 (wchar_t*) path and file names. The ASCII versions
    * throw wobblies if fed multibyte UTF-8 characters (e.g. accented ones):
    * we must convert s (and d) to UTF-16 and handle failures while doing
    * this.
    */
   wchar_t filePathW[MAX_FP_CHARS];                      // file path as UTF-16

   if (!MultiByteToWideChar(CP_UTF8, 0, s, -1, filePathW, MAX_FP_CHARS)) {
      /*
       * The conversion failed. We show an error message and exit.
       */
      std::cerr << "Error copying from " << s << std::endl;
      return 2; }

   /*
    * We successfully converted filePath to UTF-16. We try to open the source
    * file.
    */
   std::ifstream ss(filePathW, std::ios::binary);
#endif

   if (!ss.is_open()) {
      /*
       * We failed to open ss. We show an error message and exit.
       */
      std::cerr << "Error copying from ";
      perror(s);
      return 2; }

   /*
    * We construct an ofstream ds for the file named d. This is similar to
    * the procedure for ss above but ss is now open and we must close it if a
    * failure occurs when opening ds.
    */
#ifdef UNIX
   std::ofstream ds(d, std::ios::binary);

#else
   if (!MultiByteToWideChar(CP_UTF8, 0, d, -1, filePathW, MAX_FP_CHARS)) {
      std::cerr << "Error copying to " << d << std::endl;
      ss.close();
      return 2; }
   std::ofstream ds(filePathW, std::ios::binary);
#endif

   if (!ds.is_open()) {
      std::cerr << "Error copying to ";
      perror(d);
      ss.close();
      return 1; }

   /*
    * We copy the file, close the two streams and return.
    */
   ds << ss.rdbuf();
   ds.close();
   ss.close();
   return 0; }



int FileRename(const char *p, const char *n) {
   /***************************************************************************
    * This renames the file or folder whose path and name is p to path and    *
    * name n. Both p and n are UTF-8 coded. It returns 0 iff successful.      *
    ***************************************************************************/

#ifdef UNIX
   /*
    * With Linux it's easy: rename() accepts UTF-8 encoded path and file
    * names and tries to rename the file or folder.
    */
   if (rename(p, n))

#else
      /*
       * The Windows rename() only accepts ASCII coded paths and may fail if
       * fed mutibyte UTF-8 coded characters (e.g. accented ones). Instead,
       * we must convert p and d to UTF-16 (wchar_t*), check for errors, then
       * use _wrename() to remove the file.
       */
      wchar_t pPathW[MAX_FP_CHARS],
      nPathW[MAX_FP_CHARS];                                  // paths as UTF-16

   if (!MultiByteToWideChar(CP_UTF8, 0, p, -1, pPathW, MAX_FP_CHARS))
      std::cerr << "Error renaming " << p << " to " << n << std::endl;
   else if (!MultiByteToWideChar(CP_UTF8, 0, n, -1, nPathW, MAX_FP_CHARS))
      std::cerr << "Error renaming " << p << " to " << n << std::endl;
   else
      /*
       * We successfully converted the paths to UTF-16. We try to rename the
       * file or folder.
       */
      if (_wrename(pPathW, nPathW))
#endif

   {
      /*
       * The rename failed. We show an error message.
       */
      std::cerr << "Error renaming " << p << " to ";
      perror(n);
      return 2; }
   return 0; }



int FileDelete(const char *f) {
   /***************************************************************************
    * This deletes the file whose UTF-8 coded path and name is f. It returns  *
    * 0 iff successful.                                                       *
    ***************************************************************************/
#ifdef UNIX
   /*
    * With Linux it's easy: remove() accepts a UTF-8 encoded path and file
    * name and tries to delete the file.
    */
   if (remove(f))

#else
      /*
       * The Windows remove() only accepts an ASCII coded path and name and
       * may fail if fed mutibyte UTF-8 coded characters (e.g. accented
       * ones). Instead, we must convert f to UTF-16 (wchar_t*), check for
       * errors, then use _wremove() to remove the file.
       */
      wchar_t filePathW[MAX_FP_CHARS];                   // file path as UTF-16

   if (!MultiByteToWideChar(CP_UTF8, 0, f, -1, filePathW, MAX_FP_CHARS))
      std::cerr << "Error deleting " << f << std::endl;
   else
      /*
       * We successfully converted filePath to UTF-16. We try to remove the
       * file.
       */
      if (_wremove(filePathW))
#endif

   {
      /*
       * The file removal failed. We show an error message.
       */
      std::cerr << "Error deleting ";
      perror(f);
      return 2; }
   return 0; }



size_t strlcpy(char *dst, const char *src, size_t size) {
   /***************************************************************************
    * This implements the strlcpy safe copy function available in BSD. It     *
    * copies null-terminated string src to dst, truncating dst and ensuring   *
    * that it is null terminated if the length of src equals or exceeds size. *
    * The actual length of src (excluding the terminating null) is returned.  *
    * E.g.:                                                                   *
    *  if (strlcpy(dst, src, sizeof dst) >= sizeof dst)                       *
    *     printf("String is too long\n");                                     *
    ***************************************************************************/
   size_t pos = 0;                                           // position in src
   char c;                                                 // current character

   /*
    * We get successive characters from src until we have the terminating
    * null. We put up to size of them into dst.
    */
   do {
      c = src[pos];
      if (pos < size)
         dst[pos] = c;
      pos += 1; }
   while (c);

   /*
    * We decrement pos (now the count of characters in src) to exclude the
    * terminating null. If it is still >= size, we know that the terminating
    * null wasn't copied to dst. We change the last character in dst to a
    * null.
    */
   if (--pos >= size)
      dst[size - 1] = '\0';

   return pos; }

