/******************************************************************************
 * compat.h - Windows/Linux compatibility grab-bag headers. See compat.cpp.   *
 *                                                                            *
 *  Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>               *
 *                                                                            *
 *  compat is free software: you can redistribute it and/or modify it         *
 *  under the terms of the GNU General Public License as published by the     *
 *  Free Software Foundation, either version 3 of the License, or             *
 *  (at your option) any later version.                                       *
 *                                                                            *
 *  compat is distributed in the hope that it will be useful, but             *
 *  WITHOUT ANY WARRANTY; without even the implied warranty of                *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
 *  See the GNU General Public License for more details.                      *
 *                                                                            *
 *  You should have received a copy of the GNU General Public License along   *
 *  with this program.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#define MAX_FP_CHARS 500 // maximum characters in a file path

#ifdef UNIX
/*
 * With Linux, we use the available File Tree Walk function ftw(). It returns
 * file details in a standard stat struct with paths coded using UTF-8.
 *
 * We need to define kbhit() ourselves in Linux and use getchar() to fetch a
 * pressed key.
 *
 * With Linux, main() is the entry point and its arguments are UTF-8 coded.
 */
#include <ftw.h>
#define fstat stat

int fopen_s(FILE **f, const char *path, const char *mode);
int _kbhit (void);
void Sleep(int d);
#define _getch() getchar();


#else // nUNIX
/*
 * Windows needs some more header files and doesn't have ftw(). We will use
 * an equivalent defined in compat.cpp. Note that capabilites need to be
 * enabled in config.h.
 *
 * Windows uses wmain() as the entry point with wchar_t (UTF-16) coded
 * arguments.
 */
#include <windows.h>
#include <wchar.h>
#include <conio.h>

/*
 * We modify the name of main() everywhere so that Windows won't see it as a
 * possible entry point.
 */
#define main __main

/*
 * The ftw callback function.
 */
typedef int (*__ftw_func_t) (const char *__filename,
   const struct stat *__status, int __flag);

#define FTW_D   1 // __flag when a directory was found
#define FTW_F   2 // __flag when a file was found

/*
 * Partial equivalent to the Linux stat struct.
 */
struct fstat {
#ifdef ST_SIZE
   long long st_size;                                       // size of the file
#endif // ST_SIZE
#ifdef ST_ATIME
   time_t st_atime;                                         // access date/time
#endif // ST_ATIME
#ifdef ST_CTIME
   time_t st_ctime;                                       // creation date/time
#endif // ST_CTIME
#ifdef ST_MTIME
   time_t st_mtime;                                   // modification date/time
#endif // ST_MTIME
#ifdef ST_HIDDEN
   int st_hidden;                                     // TRUE if file is hidden
#endif
 };

int ftw(const char *dirpath,
   int (*fn) (const char *fpath, const struct fstat *sb,
      int typeflag),
   int nopenfd);
#endif //nUNIX
int IsHidden(const char *fpath, const struct fstat *sb);
void changemode(int dir);


/*
 * Functions to copy and delete files.
 */
int FileCopy(const char *d, const char *s);
int FileRename(const char *p, const char *n);
int FileDelete(const char *f);

/*
 * The strlcpy safe copy function from BSD.
 */
size_t strlcpy(char *dst, const char *src, size_t size);

