/******************************************************************************
 *           dj.cpp - MUSIC TRACK SELECTION UTILITY MAIN PROGRAM              *
 *                                                                            *
 * This scans a directory (including subdirectories) containing music files,  *
 * creates or updates a .CSV file which can contain ratings for each track    *
 * found and optionally creates a playlist containing tracks randomly         *
 * selected taking the ratings into account.                                  *
 *                                                                            *
 *  Copyright (C) Steve Hayes 2011-2019 <steve@honeylink.co.uk>               *
 *                                                                            *
 *  dj is free software: you can redistribute it and/or modify it             *
 *  under the terms of the GNU General Public License as published by the     *
 *  Free Software Foundation, either version 3 of the License, or             *
 *  (at your option) any later version.                                       *
 *                                                                            *
 *  dj is distributed in the hope that it will be useful, but                 *
 *  WITHOUT ANY WARRANTY; without even the implied warranty of                *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
 *  See the GNU General Public License for more details.                      *
 *                                                                            *
 *  You should have received a copy of the GNU General Public License along   *
 *  with this program.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/
#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <iostream>
#include <fstream>

#include "config.h"
#include "text.h"
#include "compat.h"
#include "function.h"
#include "options.h"
#include "trackslist.h"
#include "wordslist.h"

using namespace std;

#define FALSE 0
#define TRUE  1

#define MAX_TRACKS 0xffff    // maximum number of tracks - ideally power of 2 - 1

/*
 * Global variables (CheckFile(), ScanFile(), DeleteFile() and the TracksList
 * insertor need access to these):
 */
Options options;                                       // option switches, etc.
TracksList tracks = MAX_TRACKS;               // the list of old and new tracks
int destNum = 0;                                // used with numeric file names


int CheckFile(const char *fpath, const struct fstat *sb, int typeflag) {
   /***************************************************************************
    * This is called by ftw() when it finds a file in the music directory or  *
    * one of its subdirectories. It checks if it is a music file (3 letter    *
    * extension not .csv., .m3u or .txt). If so, it uses Set to add it to or  *
    * update its size in tracksList[]. If it's a new track, it's marked as    *
    * unrated (-1).                                                           *
    ***************************************************************************/
   static char hd[1024] = "";                  // current hidden directory path
   int i;                                                            // scratch

   if (*hd && !strStartCmp(hd, fpath))
      /*
       * This is a file or subdirectory contained in an already-identified
       * hidden directory. We ignore it.
       */
      return 0;

   if (typeflag == FTW_D && !options.includeHidden) {
      if (IsHidden(fpath, sb))
         /*
          * This is a new hidden directory. We record it.
          */
         strlcpy(hd, fpath, sizeof hd);

      else
         /*
          * This is a new non-hidden directory. We clear any previous hidden
          * directory record.
          */
         *hd = '\0'; }

   else if (typeflag == FTW_F && sb->st_size &&
      (options.includeHidden || !IsHidden(fpath, sb))) {
      /*
       * It's a file with non-zero size. If hidden files are excluded, it
       * isn't one. We see if its path matches the specified music directory.
       * If it does, we will remove it and any following slash to make the
       * path relative.
       *
       * If the size is not checked and a zero-size file is found,it gets
       * added, then purged, causing the ratings file to be rewritten every
       * time.
       */
      i = 0;
      while (options.mediaDir[i] && fpath[i] == options.mediaDir[i])
         i += 1;
      if (options.mediaDir[i])
         i = 0;                            // the music directory doesn't match
      if (i && fpath[i] == '/')
         i += 1;                                // skip over the path separator

      tracks.Set(&fpath[i], -1, NULL, options.mediaExts, sb->st_size); }

   return 0; }                                      // OK - ftw should continue


int ScanFile(const char *fpath, const struct fstat *sb, int typeflag) {
   /***************************************************************************
    * This is called by ftw() when it finds a file in the destination         *
    * directory or one of its subdirectories. It records the highest numeric  *
    * file name found as destNum.                                             *
    ***************************************************************************/
   int i = 0;
   char buf[11];

   if (typeflag == FTW_F) {
      PathParse(fpath, 0, &i);                             // get the extension
      if (MatchExt(&fpath[i], options.mediaExts)) {
         /*
          * The extension indicates it is a media file. We move to the start
          * of the filename (i.e. the end of the path) and put its name in
          * buf.
          */
         PathParse(fpath, buf, sizeof buf, 0, &i);
         if (i - 1 <= (int) strlen(options.copyDest)) {
            /*
             * The path length matches the media directory - the file isn't
             * in a subdirectory. We get its name as an integer and, if it is
             * larger than destNum, update destNum.
             */
            i = atoi(buf);
            if (i > destNum)
               destNum = i; } } }

   return 0; }                                      // OK - ftw should continue


int DeleteFile(const char *fpath, const struct fstat *sb, int typeflag) {
   /***************************************************************************
    * This is called by ftw() when it finds a file in the destination         *
    * directory or one of its subdirectories. It deletes the file if it has a *
    * media extension and isn't in a subdirectory.                            *
    ***************************************************************************/
   int pos = 0;                                            // position in fpath

   if (typeflag == FTW_F) {
      PathParse(fpath, 0, &pos);                           // get the extension
      if (MatchExt(&fpath[pos], options.mediaExts)) {
         /*
          * The extension indicates it is a media file. We move to the start
          * of the filename (i.e. the end of the path).
          */
         PathParse(fpath, 0, &pos);
         if (pos - 1 <= (int) strlen(options.copyDest))
            /*
             * The path length matches the media directory - the file isn't
             * in a subdirectory. We delete it.
             */
            FileDelete(fpath); } }

   return 0; }                                      // OK - ftw should continue

int strncmp_x(const char *a, const char *b, int n) {
   return strncmp(a, b, n); }

int main(int argc, char *argv[]) {
   /***************************************************************************
    * This is the main function for dj. Note that, with Windows it is renamed *
    * (via compat.h) and called via _tmain() in compat.cpp.                   *
    ***************************************************************************/
   int
      fChanged = FALSE,                 // TRUE if ratings file needs rewriting
      nMatches;                        // number of files found to have matches

   /*
    * The options class is global (needed by CheckFile(), etc.) and is
    * already constructed. We set options from the command line arguments.
    */
   options.Set(argv, argc);

   if (options.showHelp) {
      /*
       * The help screen is required. We display it. Strings are in text.h.
       */
      cout << "***********************************************\n";
      cout << "*                  dj                         *\n";
      cout << "*                                             *\n";
      cout << HTEXT01;                                           // Description
      cout << HTEXT02;
      cout << HTEXT03;
      cout << HTEXT04;
      cout << "*                                             *\n";
      printf(HTEXT05, "1.03");                                       // Version
      cout << "*                                             *\n";
      cout << HTEXT10;                                                // Syntax
      cout << HTEXT11;
      cout << HTEXT12;
      cout << HTEXT13;
      cout << HTEXT14;
      cout << HTEXT15;
      cout << "*                                             *\n";
      cout << HTEXT20;                                               // Options
      cout << HTEXT21;                                        // without values
      cout << HTEXT22;
      cout << HTEXT23;
      cout << HTEXT24;
      cout << HTEXT25;
      cout << HTEXT26;
      cout << HTEXT27;
      cout << HTEXT28;
      cout << HTEXT29;
      cout << HTEXT30;
      cout << "*                                             *\n";
      cout << HTEXT40;                                           // with values
      cout << HTEXT41;
      cout << HTEXT42;
      cout << HTEXT43;
      cout << HTEXT44;
      cout << HTEXT45;
      cout << HTEXT46;
      cout << HTEXT47;
      cout << HTEXT48;
      cout << HTEXT49;
      cout << HTEXT50;
      cout << HTEXT51;
      cout << HTEXT52;
      cout << HTEXT53;
      cout << HTEXT54;
      cout << "*                                             *\n";
      cout << HTEXT90;                                      // Copyright notice
      cout << HTEXT91;
      cout << HTEXT92;
      cout << "***********************************************\n"; }

   if (!*options.mediaDir)
      return 0;                                          // no point continuing

   /*
    * We try to open and read in the ratings CSV files.
    */
   {
      int i = 0,
         j;
      while (i < options.nRatingFiles) {
         ifstream in(options.ratingFile[i++]);
         if (in) {
            cout << RRFTEXT;                         // Reading ratings file...
            in >> tracks;
            in.close(); } }

      /*
       * We scan the music directory, adding missing tracks to the list. If a
       * track is already there, its size is updated.
       */
      i = tracks.GetNumber();               // tracks in existing ratings files
      cout << SMFTEXT;                              // Scanning media folder...
      ftw (options.mediaDir, &CheckFile, 10);
      cout << "  " << (j = tracks.GetNumber()) <<
         OCTTEXT <<               // old and current tracks: total current size
         tracks.GetTotalSize() << endl;
      fChanged = i != j; }                          // TRUE if new tracks added

   /*
    * We set up the root path. (This will be used when deleting or renaming
    * files and when copying selected files to a chosen directory.)
    */
   PathAndName mdir(options.mediaDir);

   if (*options.deleteAttrs) {
      /*
       * We will delete track files with any of the specified deletion
       * attributes.
       */
      cout << CTDTEXT;    // Checking for track files specified for deletion...
      int pos = 0,                                          // current position
         nDeleted = 0;                               // number of files deleted
      while (pos < tracks.GetNumber()) {
         if (tracks.GetSize(pos)
            && MatchChar(tracks.GetAttrs(pos), options.deleteAttrs)) {
            /*
             * A track was found with a specified deletion attribute. Its
             * file exists. We delete it and remove it from the list.
             */
            remove(mdir.Path(tracks.GetPath(pos)));
            tracks.Remove(pos);
            nDeleted += 1; }
         else
            pos += 1; }
      if (nDeleted) {
         cout << "  " << nDeleted << NFDTEXT;        // number of files deleted
         fChanged = TRUE; } }

   cout << CMTTEXT;            // Checking list for tracks no longer present...
   /*
    * We remove tracks from the list where the corresponding file wasn't
    * found and no rating or attribute was assigned.
    */
   { int nRemoved;                                  // number of tracks removed
      if ((nRemoved = tracks.Purge())) {
         cout << "  " << nRemoved << TRMTEXT;                 // tracks removed
         fChanged = TRUE; } }


   if (options.enableNameFix) {
      cout << FFNTEXT;                       // Fixing folder and file names...
      int nFixed = 0,                  // number of file and folder names fixed
         track = 0,                                     // current track number
         pos;                                               // position in path
      char buf[MAX_FP_CHARS],                                  // original path
         buf2[MAX_FP_CHARS],                                   // modified path
         buf3[MAX_FP_CHARS];                           // subsequent track path
      PathAndName mdir2(options.mediaDir);            // new path when renaming

      /*
       * We check each track in turn.
       */
      while (track < tracks.GetNumber()) {

         if (tracks.GetSize(track)) {
            /*
             * The track exists. We get its file path and name.
             */
            strlcpy(buf, tracks.GetPath(track), sizeof buf);
            strlcpy(buf2, buf, sizeof buf2);

            /*
             * We find the start of the file name.
             */
            pos = 0;
            PathParse(buf, 0, &pos);
            if (pos)
               PathParse(buf, 0, &pos);

            if (FixName(&buf2[pos])) {
               /*
                * The file name needs fixing. We try to rename it and, if
                * that succeeds, we update tracks.
                */
               if (!FileRename(mdir.Path(buf), mdir2.Path(buf2))) {
                  tracks.UpdatePath(track, buf2);
                  nFixed += 1; } }

            while (pos > 0) {
               /*
                * We check containing folders in turn.
                */
               int end = --pos;
               buf[end] = buf2[end] = '\0';
               if (PathParse(buf, 0, &pos)) {

                  if (FixName(&buf2[pos])) {
                     if (!FileRename(mdir.Path(buf), mdir2.Path(buf2))) {
                        /*
                         * A folder name needs fixing. We successfully
                         * renamed it. Now we must change all entries in
                         * tracks relating to that folder, including the
                         * current one.
                         */
                        nFixed += 1;
                        int track2 = tracks.GetNumber();
                        do {
                           strlcpy(buf3, tracks.GetPath(--track2), sizeof buf3);
                           if (buf3[end] == '/' && !strncmp(buf, buf3, end)) {
                              /*
                               * We found an entry in the same folder. We
                               * update it.
                               */
                              int p = strlen(buf2);
                              strlcpy(&buf2[p], &buf3[end], sizeof buf2 - p);
                              tracks.UpdatePath(track2, buf2);
                              buf2[p] = '\0'; } }                 // restore it
                        while (track2 > track); } } } } }
         track += 1; }

      if (nFixed) {
         cout << "  " << nFixed << NNCTEXT;          // number of names changed
         fChanged = TRUE; } }


   if (options.autoExclude || options.matchMode) {
      cout << MATTEXT;                                    // Matching tracks...
      int pos1 = 0;                                 // current position in list
      while (pos1 < tracks.GetNumber()) {

         if (tracks.GetSize(pos1)
            && !MatchChar(options.skipAttrs, tracks.GetAttrs(pos1))) {
            /*
             * The current track exists and doesn't have any skip matching
             * attribute(s). We find the end of its path, then look at all
             * subsequent tracks.
             */
            int ppos = 0;                       // position of filename in path
            if (PathParse(tracks.GetPath(pos1), 0, &ppos))   // extension found
               PathParse(tracks.GetPath(pos1), 0, &ppos);
            int pos2 = pos1;             // position of candidate matching file
            while (++pos2 < tracks.GetNumber()) {

               if (tracks.GetSize(pos2)
                  && !MatchChar(options.skipAttrs, tracks.GetAttrs(pos2))
                  && (options.matchSameFolder ||
                     strncmp(tracks.GetPath(pos1), tracks.GetPath(pos2), ppos))
                  && tracks.Match(pos1, pos2) > options.matchThreshold)
                  /*
                   * We found a subsequent track that exists. It doesn't have
                   * skip matching attributes. If same folder matching is
                   * disabled, the folder names don't match. Its name matches
                   * the current track's name better than the matching
                   * threshold. We link the tracks.
                   */
                  tracks.Link(pos1, pos2); } }

         pos1 += 1; }

      /*
       * We sort the list so all linked tracks appear together at its start.
       * This makes it easier to assign match numbers, easier to generate a
       * match-mode playlist and easier for the user to spot and deal with
       * matches.
       */
      if ((nMatches = tracks.SortLinks())) {
         cout << "  " << nMatches << NFMTEXT;   // number of files have matches
         if (options.matchMode)
            fChanged = TRUE; } }


   /*
    * Provided there have been changes and only one rating file was used, we
    * rewrite it so that it shows matches, includes any new (unrated) tracks
    * and omits any old unrated tracks that have since been deleted.
    */
   if (fChanged && options.nRatingFiles == 1) {
      ofstream out(options.ratingFile[0]);
      if (out) {
         cout << WRFTEXT;                            // Writing ratings file...
         out << tracks;
         out.close(); } }

   if (options.playlistTargetSize > 0) {
      /*
       * A playlist is wanted.
       */

      long long playlistSize = 0;      // total size of selected files in bytes
      int nSelectedTracks = 0;        // number of tracks selected for playlist
      int nPreviousSelected;       // number of tracks selected before the pass

      if (options.matchMode) {
         /*
          * A match mode playlist is wanted. We find out how many tracks with
          * links will fit within the requested size. (They are already in
          * position at the start of the ordered list.)
          */
         do {
            playlistSize += tracks.GetSize(nSelectedTracks++); }
         while (nSelectedTracks < nMatches
            && playlistSize < options.playlistTargetSize);
         nSelectedTracks -= 1; }

      else {
         /*
          * A normal playlist is wanted. We seed the random generator (to get
          * a different playlist each time), set the default rating for
          * unrated tracks and shuffle the list.
          */
         cout << SELTEXT;                                // Selecting tracks...
         srand((unsigned int) time(NULL));
//         srand(0); // uncomment for repeatability when testing
         tracks.DefaultRate(options.initialRating);
         tracks.Shuffle();

         int avgSelectedRating = 0;         // average rating of playlist files
         do {
            long long size;                 // size in bytes of candidate track
            int rating;                            // rating of candidate track
            /*
             * Tracks from positions 0 to nSelectedTracks-1 in the list are
             * ones already selected for the playlist.
             *
             * We loop until we don't add any tracks during a pass. Usually
             * this is because the playlist is nearly full.
             *
             * (If there aren't many tracks, there's a significant chance of
             * not selecting any during a pass. In that situation, we
             * continue until after the pass that has a ratingWeighting of
             * zero. )
             *
             * We look at unselected candidate tracks.
             */
            nPreviousSelected = nSelectedTracks;
            for (int pos = nSelectedTracks; pos < tracks.GetNumber(); pos++) {
               /*
                * We check if the candidate track meets all of the following
                * requirements:
                *  Its file exists (non-zero size)
                *  It can be added without making the playlist too big
                *  Its rating is not too high or too low
                *  If Include attributes are set, it has at least one of them
                *  It doesn't have any Exclude attributes
                *  If Auto-exclude is enabled, it isn't a linked track
                *  Its rating, weighting and a random value say to use it
                */
               if ((size = tracks.GetSize(pos))
                  && size < options.playlistTargetSize - playlistSize
                  && (rating = tracks.GetRating(pos)) >= options.minRating
                  && rating <= options.maxRating
                  && (!*options.includeAttrs
                     || MatchChar(tracks.GetAttrs(pos), options.includeAttrs))
                  && !MatchChar(tracks.GetAttrs(pos), options.excludeAttrs)
                  && (!options.autoExclude || tracks.FirstLink(pos))
                  && tracks.UseTrack(pos, options.ratingWeighting)) {
                  /*
                   * We want to use the track. We add its size and rating,
                   * swap it with the first unselected track in the sequence
                   * and increment nSelectedTracks.
                   */
                  playlistSize += size;
                  avgSelectedRating += rating;
                  tracks.Swap(pos, nSelectedTracks++); } } }
         while ((nSelectedTracks > nPreviousSelected || tracks.GetNumber() < 500)
            && options.ratingWeighting--);

         /*
          * We calculate the average rating of the new playlist and show this
          * along with the number of tracks selected.
          */
         avgSelectedRating = nSelectedTracks ?
            (avgSelectedRating << 8) / nSelectedTracks : 0;
         printf(TSRTEXT,                     // tracks selected: average rating
            nSelectedTracks, avgSelectedRating >> 8,
            ((avgSelectedRating & 0xff) * 100) / 256);

         cout << LPLTEXT;                              // Levelling playlist...
         /*
          * We shuffle the selected tracks to overcome the tendency of the
          * last ones to have lower ratings.
          */
         tracks.Shuffle(nSelectedTracks);

         /*
          * We scan the playlist to break up any runs of tracks with ratings
          * above or below the average.
          */
         {
            int pos1 = 0,                       // current position in playlist
               pos2 = 0,                    // position of candidate next track
               cr = 0,                                     // cumulative rating
               ncr,                                   // next cumulative rating
               rating,                                     // swap track rating
               nSwaps = 0;                          // number of tracks swapped
            while (pos1 < nSelectedTracks) {
               /*
                * We loop until we reach the end of the list. We calculate
                * the next cumulative rating (keeping tracks in the current
                * order).
                */
               ncr = cr + (tracks.GetRating(pos2 = pos1) << 8)
                  - avgSelectedRating;
               if (cr > 0 && ncr > options.variationThreshold
                  || cr < 0 && ncr < -options.variationThreshold)
                  /*
                   * Either the cumulative rating is high and this track
                   * would push it above the limit or it is low and the track
                   * would push it below the limit. If the rating is too
                   * high, we look for a below-average track, otherwise we
                   * look for an above-average one.
                   */
                  while (++pos2 < nSelectedTracks &&
                     (((rating = tracks.GetRating(pos2) << 8)
                           > avgSelectedRating) ? cr > 0 : cr < 0));

               if (pos2 > pos1 && pos2 < nSelectedTracks) {

                  /*
                   * We need to modify the playlist to level it. We swap the
                   * next track with the more suitable one, count swaps and
                   * update the cumulative rating.
                   */
                  tracks.Swap(pos1, pos2);
                  nSwaps += 1;
                  cr = cr + rating - avgSelectedRating; }

               else
                  cr = ncr;                                 // using next track

               /*
                * We adjust the cumulative rating to counter any persistent
                * rounding error, then prepare for the following track.
                */
               cr += (cr >= 0) ? -1 : 1;
               pos1 += 1; }

            if (nSwaps)
               cout << "  " << nSwaps << TNSTEXT; } } // number of tracks swapped

      if (*options.playlistFile) {
         cout << WPFTEXT;                           // Writing playlist file...
         /*
          * We open the output playlist file to replace or append according
          * to options.
          */
         ofstream out(options.playlistFile,
            options.keepExisting ? ios_base::app : ios_base::trunc);

         if (out) {
            /*
             * We opened it successfully. We write the selected track paths
             * and names to it.
             */
            int pos = 0;
            while (pos < nSelectedTracks)
               out << tracks.GetPath(pos++) << endl;
            out.close(); }
         else {
            /*
             * We couldn't open the file. We show an error message.
             */
            cerr << OPETEXT;                              // Error: cannot open
            perror(options.playlistFile); } }

      if (*options.copyDest) {
         /*
          * We set up the destination path and open a playlist file there to
          * replace or append according to options.
          */
         PathAndName ddir(options.copyDest);          // set destination folder
         ofstream out(*options.copyPlaylistFile ?
            options.copyPlaylistFile : ddir.Path("dj.m3u"),
            options.keepExisting ? ios_base::app : ios_base::trunc);


         if (options.keepExisting) {
            /*
             * We want to copy the selected files to a folder or device,
             * keeping any existing ones too. If we are writing as numeric
             * file names, we need to scan the existing files so we can
             * continue their numbering sequence.
             */
            if (options.numericNames) {
               cout << SCNTEXT;             // Scanning existing track files...
               ftw (options.copyDest, &ScanFile, 10);
               if (destNum)
                  cout << FNFTEXT            // First new file will be numbered
                  << destNum + 1 << endl; } }

         else {
            cout << DELTEXT;                // Deleting existing track files...
            ftw (options.copyDest, &DeleteFile, 10); }

         cout << CPYTEXT;                             // Copying track files...
         int pos = 0;
         while (pos < nSelectedTracks) {
            int ppos = 0;                                   // position in path

            if (options.numericNames) {
               if (PathParse(tracks.GetPath(pos), 0, &ppos)) // got file extension
                  ppos -= 1;                              // to include the dot

               /*
                * We copy the file. The destination name consists of the next
                * number plus the existing extension.
                */
               FileCopy(ddir.Path(++destNum, &tracks.GetPath(pos)[ppos]),
                  mdir.Path(tracks.GetPath(pos))); }

            else {
               PathParse(tracks.GetPath(pos), 0, &ppos);  // get file extension
               PathParse(tracks.GetPath(pos), 0, &ppos);       // get file name

               /*
                * We copy the file. The destination name consists of the
                * source name and extension.
                */
               FileCopy(ddir.Path(&tracks.GetPath(pos)[ppos]),
                  mdir.Path(tracks.GetPath(pos))); }

            if (out)
               /*
                * The output folder playlist file is open. We add the newly
                * copied file to it.
                */
               out << ddir.GetFile() << endl;
            pos += 1; }

         if (out)
            /*
             * We close the playlist file.
             */
            out.close(); } }


   cout << DONETEXT;  // Done - Visit drenewydd.com for support and information
   return 0; }
