/******************************************************************************
 * function.cpp - Miscellaneous grab-bag functions.                           *
 *                                                                            *
 * Platform-independent functions are kept here. See also function.h.         *
 *                                                                            *
 *  Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>               *
 *                                                                            *
 *  This is free software: you can redistribute it and/or modify it           *
 *  under the terms of the GNU General Public License as published by the     *
 *  Free Software Foundation, either version 3 of the License, or             *
 *  (at your option) any later version.                                       *
 *                                                                            *
 *  It is distributed in the hope that it will be useful, but                 *
 *  WITHOUT ANY WARRANTY; without even the implied warranty of                *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
 *  See the GNU General Public License for more details.                      *
 *                                                                            *
 *  You should have received a copy of the GNU General Public License along   *
 *  with this program.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/
#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sstream>
#include <iomanip>

#include <iostream>


#include "config.h"
#include "compat.h"
#include "function.h"
#include "text.h"



char opt(char *args[], int maxarg, int *num, int *pos, const char *valopts,
   char *val, int valsize) {
   /***************************************************************************
    * This gets successive options, values and parameters from command line   *
    * arguments args[] (e.g. argv[]). maxarg is the number of arguments (e.g. *
    * argc).                                                                  *
    *                                                                         *
    * *num should be initialised to 1 (to start with argv[1]) and *pos should *
    * be initialised to 0. valopts is a string of option letters where a      *
    * value is expected to follow. opt is then called repeatedly.             *
    *                                                                         *
    * Where an argument starts with -, one or more single character options   *
    * are expected to follow and are returned one at a time in succession.    *
    *                                                                         *
    * When an option character is found that is included in valopts, a value  *
    * is expected to follow, either immediately in the same argument or as    *
    * the following argument. The option is returned with its value placed in *
    * val. valsize should be sizeof val to guard against buffer overflows.    *
    *                                                                         *
    * Possible return conditions are:                                         *
    *                                                                         *
    * Option character returned with zero-length val - the option is not in   *
    * valopts or it is but no value follows immediately and there are no more *
    * arguments or the next argument starts with -.                           *
    *                                                                         *
    * Option character returned, val not zero length - the option is in       *
    * valopts and a value was found.                                          *
    *                                                                         *
    * NULL returned, val not zero length - a parameter (not part of an        *
    * option) was found and returned as val.                                  *
    *                                                                         *
    * NULL returned and val is zero-length - there are no more options or     *
    * parameters.                                                             *
    *                                                                         *
    * Example where valopts is "wxyz" and parameters are                      *
    *  -ab 123 -cw456 -x -y 789 0 -z                                          *
    *                                                                         *
    * Successive return values are:                                           *
    *  'a',  zero-length val   (no value expected)                            *
    *  'b',  zero-length val   (no value expected)                            *
    *  NULL, val is "123"      (a parameter instead of an option)             *
    *  'c',  zero-length val   (no value expected)                            *
    *  'w',  val is "456"      (value expected and found)                     *
    *  'x',  zero-length val   (value not found)                              *
    *  'y',  val is "789"      (value expected and found)                     *
    *  NULL, val is "0"        (another parameter)                            *
    *  'z',  zero-length val   (value not found)                              *
    *  NULL, zero-length val   (no more options or parameters)                *
    ***************************************************************************/

   char optchar = 0;                                        // option character

   *val = '\0';                                       // in case no value found

   while (1) {
      if (*num < maxarg && !args[*num][*pos]) {
         /*
          * We finished processing the current argument. We move on to the
          * next one.
          */
         *num += 1;
         *pos = 0; }

      if (*num >= maxarg)
         /*
          * There are no more arguments. We are done.
          */
         return 0;

      if (!*pos && args[*num][0] == '-')
         /*
          * We are starting an option argument. We skip over the -.
          */
         *pos = 1;

      if (*pos) {
         if (optchar)
            /*
             * We found an option which we expected to include a value. None
             * was found, just another option. We return the first option.
             */
            return optchar; }

      else {
         /*
          * We found an argument not starting with - (a value). We copy it
          * and return.
          */
         strlcpy(val, args[(*num)++], valsize);
         return optchar; }

      if ((optchar = args[*num][(*pos)++])) {
         /*
          * We got an option. We check if it is one where a value is
          * expected.
          */
         int pos2 = 0;
         while (valopts[pos2] && valopts[pos2] != optchar)
            pos2 += 1;

         if (!valopts[pos2])
            /*
             * No value is expected. We are done.
             */
            return optchar;

         if (args[*num][*pos]) {
            /*
             * There is a value immediately after the option.
             */
            strlcpy(val, &args[*num][*pos], valsize);
            *num += 1;
            *pos = 0;
            return optchar; } } } }


int UserConfirm() {
   /***************************************************************************
    * This asks the user to confirm an action that cannot be undone. It       *
    * returns TRUE if they answer Y.                                          *
    ***************************************************************************/
   char c;                                           // the key that they press

   changemode(1);                                            // needed for UNIX
   std::cout << UCFTEXT;           // This can't be undone! Are you sure (y/N)?

   /*
    * We wait for a keypress, then get the key.
    */
   while (!_kbhit()) Sleep(200);
   c = _getch();
   changemode(0);

   return c == YESCHAR || c == YESCHAR + 'a' - 'A'; }



char *PathParse(const char *path, char *dest, int destsize, int word, int *pos) {
   /***************************************************************************
    * This is used to parse path (a file path and name) from end to           *
    * beginning, returning individual parts such as the extension, the file   *
    * name or a containing directory. It can also return individual words in  *
    * the file name. The selected part is returned in dest: destsize should   *
    * be sizeof dest to guard against buffer overflows. The return value is   *
    * normally dest.                                                          *
    *                                                                         *
    * If *pos is zero, the file extension is returned and *pos is set to      *
    * point to its start. This should normally be the first call.             *
    *                                                                         *
    * When *pos is nonzero and word is FALSE (0), the file name or a          *
    * containing directory is returned and *pos is set to point to its start. *
    *                                                                         *
    * When *pos is nonzero and word is TRUE (nonzero), a word from the file   *
    * name or containing directory name is returned and *pos is set to point  *
    * to its start.  When word is > 0, NULL is returned if *pos is already    *
    * pointing to the first word in the file or directory name.               *
    *                                                                         *
    * The calling code should check if *pos is zero indicating that parsing   *
    * is complete. If word was > 0, it should also check if NULL is returned. *
    * This indicates that there are no more words in the name: PathParse can  *
    * be called again with word < 0 to start getting words from the           *
    * containing directory name.                                              *
    *                                                                         *
    * Note that a zero-length dest may be returned if path is zero length, if *
    * there is no file extension or if there are successive non-alphabetic    *
    * characters.                                                             *
    ***************************************************************************/
   int len;                                               // length of the part

   if ((len = PathParse(path, word, pos)) < 0)
      return NULL;
   if (len>= --destsize)
      /*
       * The field (part of path) is too long to fit in dest. We will
       * truncate it.
       */
      len = destsize;

   /*
    * We set the end-of-string in dest, copy the field to dest and return it.
    */
   dest[len] = '\0';
   while (len) {
      len -= 1;
      dest[len] = path[len + *pos]; }

   return dest; }



int PathParse(const char *path, int word, int *pos) {
   /***************************************************************************
    * This is used by the previous PathParse to find the start of the next    *
    * part of path. It sets *pos to point to it and returns the length of the *
    * part.                                                                   *
    *                                                                         *
    * If *pos is 0, it looks for a file extension and sets *pos to point to   *
    * it. If no . is found before a path separator (/, \ or :), it sets *pos  *
    * to the length of path plus 1 and returns zero.                          *
    *                                                                         *
    * Otherwise if word is FALSE (0), it looks for a path separator before    *
    * the part. If word is TRUE (nonzero), it looks for any non-alphabetic    *
    * character. When word is positive, it first checks to see if it is at a  *
    * path separator and, if so, it returns -1. Otherwise, it returns the     *
    * length of the part found.                                               *
    ***************************************************************************/
   int end,                   // usually the end of the part (our initial *pos)
      cpos;                                                 // current position
   char cchar = '\0';                                      // current character

   if (!(cpos = end = *pos)) {
      /*
       * We are just starting and will look for the file extension. We move
       * to the end of path;
       */
      if (!(cpos = end = strlen(path)))
         /*
          * path is zero-length. We return immediately.
          */
         return 0;

      /*
       * We work backwards, looking for . or a path separator.
       */
      while (cpos && cchar != '.' && cchar != '/'
#ifndef UNIX
         && cchar != '\\' && cchar != ':'
#endif // nUNIX
         )
      cchar = path[--cpos];

      if (cchar != '.') {
         /*
          * No extension was found. We set *pos ready to get the file name or
          * last word in it, then return.
          */
         *pos = end;
         return 0; } }

   else {
      /*
       * We are currently just after . or the null at the end of path or a
       * path or word separator. We move back to its position.
       */
      cpos = end -= 1;

      if (word) {
         /*
          * We are looking for an individual word in a file name.
          */
         cchar = path[cpos];

         if (word > 0 && (cchar == '/'
#ifndef UNIX
               || cchar == '\\' || cchar == ':'
#endif // nUNIX
               ))
         /*
          * We are at a path separator: we have finished the file name.
          */
         return -1;

         /*
          *  We search backwards until we
          * find a non-alphabetic character or pass the start of path.
          */
         cchar = 'A';                                           // a valid char
         while (cpos && (cchar >= 'A' && cchar <= 'Z'
               || cchar >= 'a' && cchar <= 'z'))
            cchar = path[--cpos]; }

      else {
         /*
          * We are looking for the file name or a directory name. We search
          * backwards until we find a path separator or pass the start of
          * path.
          */
         while (cpos && cchar != '/'
#ifndef UNIX
            && cchar != ':'
#endif // nUNIX
            )
         cchar = path[--cpos]; } }

   /*
    * We set *pos to the start of the part and find and return its length.
    */
   *pos = cpos = (cpos > 0) ? cpos + 1 : 0;
   end -= cpos;
   return end; }



CSVLine::CSVLine(const char *line) {
   /***************************************************************************
    * CONSTRUCTOR CSVLine: this copies Comma Separated Values line line into  *
    * lbuf, splits it into fields, counts them as nFields and records their   *
    * starting positions in fpos[].                                           *
    *                                                                         *
    * Inline methods that return the fields are in function.h.                *
    ***************************************************************************/
   int len,                                                       // sizeof buf
      pos,                                          // current position in line
      endPos;                                  // end position of current field

   /*
    * We create the buffer and copy line into it.
    */
   buf = new char[len = strlen(line) + 1];
   strlcpy(buf, line, len);

   nFields = 0;
   pos = 0;
   while (buf[pos] && nFields < MAX_CSVFIELDS) {
      /*
       * We look for a field. We start by skipping over any newlines or white
       * space.
       */
      while (buf[pos] == '\n' || buf[pos] == ' ' || buf[pos] == '\t')
         pos += 1;

      if (buf[pos] == '\"') {
         /*
          * The field is quoted. We skip over the opening quotes and any
          * additional leading white space and record the beginning position.
          */
         pos += 1;
         while (buf[pos] == ' ' || buf[pos] == '\t')
            pos += 1;
         fpos[nFields++] = endPos = pos;

         /*
          * We look for the closing quotes (or newline or end of buf) and
          * record one past the position of the last non-white-space
          * character before it.
          */
         while (buf[pos] && buf[pos] != '\"' && buf[pos] != '\n') {
            if (buf[pos++] > ' ')
               endPos = pos; }                   // buf[pos] wasn't white space

         if (buf[pos])
            /*
             * A closing quote or newline was found. We skip over it.
             */
            pos += 1;

         /*
          * We look for a field separator (comma or semicolon) or the end of
          * buf.
          */
         while (buf[pos] && buf[pos] != ',' && buf[pos] != ';')
            pos += 1; }

      else if (buf[pos]) {
         /*
          * The field is not quoted. We record the beginning position.
          */
         fpos[nFields++] = endPos = pos;

         /*
          * We look for a field separator (comma or semicolon) or a newline
          * or the end of line. We record one past the position of the last
          * non-white-space character before it.
          */
         while (buf[pos] && buf[pos] != '\n'
            && buf[pos] != ',' && buf[pos] != ';') {
            if (buf[pos++] > ' ')
               endPos = pos; } }                 // buf[pos] wasn't white space

      if (buf[pos])
         /*
          * A field separator or newline was found. We skip over it.
          */
         pos += 1;

      /*
       * We mark the end of the newly split field (before any trailing white
       * space, quote or field separator).
       */
      buf[endPos] = '\0';
   } }



PathAndName::PathAndName(const char *path) {
   /***************************************************************************
    * CONSTRUCTOR PathAndName - sets up root path, adding a path separator if *
    * needed.                                                                 *
    ***************************************************************************/

   strlcpy(buf, path, sizeof buf - 1);
   pos = strlen(buf);
   if (pos && buf[pos - 1] != '/')
      buf[pos++] = '/'; }



char *PathAndName::Path(const char *name) {
   /***************************************************************************
    * This appends name to the root path and returns the result.              *
    ***************************************************************************/

   strlcpy(&buf[pos], name, sizeof buf - pos);
   return buf; }



char *PathAndName::Path(int num, const char *ext) {
   /***************************************************************************
    * This appends a file name consisting of num (as a string) and extension  *
    * ext to the root path and returns the result. ext must include the dot.  *
    ***************************************************************************/
   std::ostringstream snum;
   int pos2;
   snum << std::setw(6) << std::setfill('0') << num;
   strlcpy(&buf[pos], snum.str().c_str(), sizeof buf - pos);    // numeric name
   pos2 = pos + strlen(&buf[pos]);
   strlcpy(&buf[pos2], ext, sizeof buf - pos2);
   return buf; }



char *PathAndName::GetFile() {
   /***************************************************************************
    * This returns the file name and extension portion of the path.           *
    ***************************************************************************/
   return &buf[pos]; }



int FixName(char *s) {
   /***************************************************************************
    * This checks s for characters which can cause problems in FAT folder or  *
    * file names. If it finds them, it changes them as shown:                 *
    *  " ? or : are removed                                                   *
    *  * \ | + space are changed to _                                         *
    *  < [ are changed to (                                                   *
    *  > ] are changed to )                                                   *
    *                                                                         *
    * TRUE is returned if changes are made, FALSE if not.                     *
    ***************************************************************************/
   int pos1 = 0,                                       // current position in s
      pos2 = 0,                                    // destination position in s
      r = 0;                             // result - TRUE once a change is made
   char c;                                                 // current character

   do {
      /*
       * We get a character. If it is ", ? or : we mark a change, then ignore
       * it.
       */
      c = s[pos1++];
      if (c == '\"' || c == '?' || c == ':')
         r = 1;

      else {
         /*
          * We will use the character but may need to change it.
          */
         if (c == ' ' || c == '|' || c == '\\' || c == '*' || c == '+') {
            r = 1;
            c = '_'; }
         else if (c == '<' || c == '[') {
            r = 1;
            c = '('; }
         else if (c == '>' || c == ']') {
            r = 1;
            c = ')'; }

         if (r)
            /*
             * Changes have been made: we write this character.
             */
            s[pos2] = c;
         pos2 += 1; } }
   while (c);
   return r; }



char *UnixPath(char *p) {
   /***************************************************************************
    * This replaces any backslashes in p with forward slashes (Windows only). *
    * It returns p for convenience.                                           *
    ***************************************************************************/
#ifndef UNIX
   int pos = 0;

   while (p[pos]) {
      if (p[pos] == '\\')
         p[pos] = '/';
      pos += 1; }
#endif

   return p; }



char *EscapePath(char *d, const char *s, int ds, int xl) {
   /***************************************************************************
    * This copies s to d, changing characters that could cause problems in a  *
    * csv field to escape sequences. ds is sizeof d (to guard against buffer  *
    * overruns). Fewer characters are escaped when xl is TRUE: Excel handles  *
    * CSV better than LibreOffice or Polaris Office.                          *
    *                                                                         *
    * When xl is TRUE, \ is converted to \\ and " to \2. When xl is FALSE     *
    * also , is converted to \c, ; to \s, TAB to \t, ' to \1 and SPACE to \_. *
    *                                                                         *
    * For convenience, d is returned.                                         *
    ***************************************************************************/
   int dp = 0,                                                 // position in d
      sp = 0,                                                  // position in s
      p;                                                             // scratch
   char c;                                                 // current character

   ds -= 1;                                  // leave room for terminating null
   while ((c = s[sp++])) {
      p = 0;
      while (p < 7 && c != "\\\"\',;\t "[p])
         p += 1;

      if (p < 2 || p < 7 && !xl)
         /*
          * We got a problem character. We insert a \ and translate the
          * character.
          */
         if (dp < ds) {
         d[dp++] = '\\';
         c = "\\21cst_"[p]; }

      if (dp < ds)
         /*
          * There is room in the buffer. We save the character.
          */
         d[dp++] = c; }

   /*
    * We terminate d and return it.
    */
   d[dp] = '\0';
   return d; }



char *UnEscapePath(char *d, const char *s) {
   /***************************************************************************
    * This copies s to d, undoing any changes that EscapePath() above may     *
    * have made. For convenience, d is returned.                              *
    *                                                                         *
    * IMPORTANT: there isn't a buffer overrun check. Ensure d is long enough  *
    * to hold s unchanged.                                                    *
    ***************************************************************************/
   int dp = 0,                                                 // position in d
      sp = 0,                                                  // position in s
      p;                                                             // scratch
   char c,                                                 // current character
      c2;                                // second character in escape sequence

   if (s) {
      while ((c = s[sp++])) {
         if (c == '\\') {
            /*
             * We may have an escape sequence. We get the second character
             * and check it.
             */
            c2 = s[sp];
            p = 0;
            while (p < 7 && c2 != "\\21cst_"[p])
               p += 1;

            if (p < 7) {
               /*
                * We got a valid sequence. We translate it and absorb the
                * second character.
                */
               c = "\\\"\',;\t "[p];
               sp += 1; } }

         d[dp++] = c; } }                                 // save the character

   /*
    * We terminate d and return it.
    */
   d[dp] = '\0';
   return d; }



int MatchChar(const char *a, const char *b) {
   /***************************************************************************
    * This returns TRUE (1) iff there is a character in a that is also in b.  *
    ***************************************************************************/
   int apos = 0,                                               // position in a
      bpos;                                                    // position in b
   char achar,                                                   // char from a
      bchar;                                                     // char from b

   while ((achar = a[apos++])) {
      bpos = 0;
      while ((bchar = b[bpos++])) {
         if (achar == bchar)
            return 1; } }
   return 0; }



int MatchExt(const char *a, const char *b) {
   /***************************************************************************
    * This returns TRUE (1) iff string a matches an entry in comma-separated  *
    * list b.                                                                 *
    ***************************************************************************/
   int apos = 0,                                               // position in a
      bpos = 0,                                                // position in b
      match = 1;                               // FALSE if string doesn't match
   char bchar;                                                   // char from b

   do {
      bchar = b[bpos++];

      if (bchar == ',' || !bchar) {
         /*
          * We reached a , or the end of b. If we have reached the end of a
          * and match is still TRUE, we matched an entry.
          */
         if (match && !a[apos])
            return 1;

         /*
          * We didn't match the entry. We prepare for the next one.
          */
         apos = 0;
         match = 1; }

      else if (bchar != a[apos++])
         match = 0; }                            // the character doesn't match
   while (bchar);

   /*
    * We reached the end of b without finding a match.
    */
   return 0; }


int strStartCmp(const char *a, const char *b) {
   /***************************************************************************
    * This returns FALSE if b starts with a.                                  *
    ***************************************************************************/
   int p = 0;
   while (a[p] && a[p] == b[p])
      p += 1;
   return a[p] != '\0'; }


unsigned int RSHash(const char* str) {
   /***************************************************************************
    * Robert Sedgwick's "Algorithms in C" hash function. It returns a hash    *
    * value for null-terminated str.                                          *
    ***************************************************************************/
   unsigned int
      b    = 378551,
      a    = 63689,
      hash = 0,
      i    = 0,
      len;

   len = strlen(str);
   for (i = 0; i < len; str++, i++) {
      hash = hash * a + (*str);
      a    = a * b; }

   return hash; }
