/******************************************************************************
 * options.h - Class with inline methods that parses command line arguments   *
 * and sets up options for dj accordingly.                                    *
 *                                                                            *
 *  Copyright (C) Steve Hayes 2012-2013 <steve@honeylink.co.uk>               *
 *                                                                            *
 * This program is free software; you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published by the Free *
 * Software Foundation; either version 2 of the License, or (at your option)  *
 * any later version.                                                         *
 *                                                                            *
 * This program is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with this program. If not, see <http://www.gnu.org/licenses/>.             *
 ******************************************************************************/
#define MAX_RATING_FILES 10
#define MAX_RATING 6

class Options {
   public:
   long long
      playlistTargetSize;

   int autoExclude,
      showHelp,
      excelMode,
      includeHidden,
      enableNameFix,
      matchMode,
      enableLevelling,
      matchSameFolder,
      numericNames,
      keepExisting,
      interactive,
      ratingWeighting,
      nRatingFiles,
      initialRating,
      minRating,
      maxRating,
      matchThreshold,
      variationThreshold;

   char mediaDir[100],
      ratingFile[MAX_RATING_FILES][100],
      copyDest[100],
      copyPlaylistFile[100],
      playlistFile[100],
      mediaExts[100],
      deleteAttrs[11],
      skipAttrs[11],
      includeAttrs[27],
      excludeAttrs[27];

Options() {
      /************************************************************************
       * CONSTRUCTOR - This sets default values.                              *
       ************************************************************************/
      playlistTargetSize = 0;
      autoExclude = 0;
      showHelp = 0;
//      enableMatch = 0;
      excelMode = 0;
      includeHidden = 0;
      enableNameFix = 0;
      matchMode = 0;
      matchSameFolder = 0;
      numericNames = 0;
      nRatingFiles = 0;
      keepExisting = 0;
      interactive = 1;
      ratingWeighting = 5;
      initialRating = 3;
      minRating = -1;
      maxRating = MAX_RATING;
      matchThreshold = 15;
      variationThreshold = 512;

      *mediaDir = '\0';
      *copyDest = '\0';
      *copyPlaylistFile = '\0';
      *playlistFile = '\0';
      strlcpy(mediaExts, "aac,aiff,alac,flac,mp3,m4a,m4p,ogg,wav,wma",
         sizeof mediaExts);
      *deleteAttrs = '\0';
      strlcpy(skipAttrs, "DES", sizeof skipAttrs);
      *excludeAttrs = '\x1';         // Default set later: depends on matchMode
      *includeAttrs = '\0'; }


void Set(char *args[], int maxarg) {
      /************************************************************************
       * This sets members according to command line options passed to it as  *
       * args[] (argv) and maxarg (argc).                                     *
       *                                                                      *
       * It would be tidier if this could be done in the constructor but we   *
       * need to construct options globally and argv and argc are only        *
       * available in main().                                                 *
       ************************************************************************/
      char optVal[100],                                         // option value
         optLet;                                               // option letter

      int rawArgNum = 1,             // raw argument on command line (e.g. -an)
         rawArgPos = 0,                         // position within raw argument
         argNum = 0,              // logical argument number (excludes options)
         ratingFileNum = 0;                          // position in ratingFiles


      while ((optLet = opt(args, maxarg, &rawArgNum, &rawArgPos,
               "bcdeiloprstuvwx", optVal, sizeof optVal))
         || *optVal) {

         switch (optLet) {

            case (int) NULL:
               /*
                * We got a logical argument (not an option). We process it.
                */
               switch (argNum++) {

               case 0:
                  /*
                   * We get the music directory (default "."), change any
                   * Windows \s to /s and remove any trailing /.
                   */
                  strlcpy(mediaDir, optVal, sizeof mediaDir);
#ifndef UNIX
                  UnixPath(mediaDir);
#endif // nUNIX
                  { int len = strlen(mediaDir);
                     if (len > 0 && mediaDir[len - 1] == '/')
                        mediaDir[len - 1] = '\0'; }
                  break;

               case 1:
                  /*
                   * We get the wanted playlist size, applying any applicable
                   * multiplier (K, M or G).
                   */
                  { int pos = strlen(optVal);       // position in the argument
                     int mult = 1;                                // multiplier
                     if (pos > 0) {
                        switch (optVal[pos - 1]) {
                        case 'k':
                        case 'K':
                           mult = 1024;
                           break;

                        case 'm':
                        case 'M':
                           mult = 1048576;
                           break;

                        case 'g':
                        case 'G':
                           mult = 1073741824; } }


                  if (mult > 1)
                     optVal[pos - 1] = '\0';
                  playlistTargetSize = (long long) mult
                     * (long long) atoi(optVal); } }
            break;

            /*
             * Options without values:
             */
            case 'a':
               autoExclude = 1;
               break;

            case 'f':
               enableNameFix = 1;
               break;

            case 'g':
               excelMode = 1;
               break;

            case 'h':
               showHelp = 1;
               break;

            case 'k':
               keepExisting = 1;
               break;

            case 'm':
               matchMode = 1;
               break;

            case 'n':
               numericNames = 1;
               break;

            case 'q':
               matchSameFolder = 1;
               break;

            case 'y':
               interactive = 0;
               break;

            case 'z':
               includeHidden = 1;
               break;

            /*
             * Options with values:
             */
            case 'b':
               strlcpy(copyPlaylistFile, optVal, sizeof copyPlaylistFile);
               break;

            case 'c':
               /*
                * We set copyDest, change any Windows \s to /s and remove any
                * trailing /.
                */
               strlcpy(copyDest, optVal, sizeof copyDest);
#ifndef UNIX
               UnixPath(copyDest);
#endif // nUNIX
               { int len = strlen(copyDest);
                  if (len > 0 && copyDest[len - 1] == '/')
                     copyDest[len - 1] = '\0'; }
               break;

            case 'd':
               strlcpy(deleteAttrs, optVal, sizeof deleteAttrs);
               break;

            case 'e':
               strlcpy(mediaExts, optVal, sizeof mediaExts);
               break;

            case 'i':
               initialRating = atoi(optVal);
               break;

            case 'l':
               minRating = atoi(optVal);
               break;

            case 'o':
               strlcpy(includeAttrs, optVal, sizeof includeAttrs);
               break;

            case 'p':
               strlcpy(playlistFile, optVal, sizeof playlistFile);
               break;

            case 'r':
               if (nRatingFiles < MAX_RATING_FILES)
                  strlcpy(ratingFile[nRatingFiles++], optVal, sizeof ratingFile[0]);
               break;

            case 's':
               strlcpy(skipAttrs, optVal, sizeof skipAttrs);
               break;

            case 't':
               matchThreshold = atoi(optVal);
               break;

            case 'u':
               maxRating = atoi(optVal);
               break;

            case 'v':
               variationThreshold = atoi(optVal) << 8;
               break;

            case 'w':
               ratingWeighting = atoi(optVal);
               break;

            case 'x':
               strlcpy(excludeAttrs, optVal, sizeof excludeAttrs); } }

      /*
       * We change some default values according to other specified options.
       */
      if (!*mediaDir)
         showHelp = 1;

      if (!nRatingFiles)
         strlcpy(ratingFile[nRatingFiles++], "dj.csv", sizeof ratingFile[0]);

      if (minRating < 0)
         minRating = matchMode ? 0 : 1;

      if (playlistTargetSize && !*playlistFile) {
         /*
          * We set the default playlist path and file name if none was
          * specified and a playlist is requested.
          */
         strlcpy(playlistFile, mediaDir, sizeof playlistFile);
         int pos = strlen(playlistFile);
         if (pos < sizeof playlistFile - 2) {
            playlistFile[pos++] = '/';
            strlcpy(&playlistFile[pos], "dj.m3u",
               sizeof playlistFile - pos); } }

      if (*excludeAttrs == '\x1')
         strlcpy(excludeAttrs, matchMode ? "" : "DEX", sizeof excludeAttrs);

      if (*mediaDir && interactive) {
         /*
          * We ask the user to confirm any actions that will irreversably
          * change their media files.
          */
         if (*deleteAttrs) {
            printf (
               CFDTEXT      // You ask to delete media files with attribute set
               , deleteAttrs);
            if (!UserConfirm())
            *deleteAttrs = '\0'; }                         // oops, don't do it

         if (enableNameFix) {
            std::cout <<
               CNFTEXT;           // You ask to fix FAT-incompatible filenames;
            if (!UserConfirm())
            enableNameFix = 0; } } }                       // oops, don't do it

 };                                                              // END Options
