/******************************************************************************
 * trackslist.cpp - Class that maintains a list of tracks with file path and  *
 * name, rating, attributes, size, words from the file name and matches. See  *
 * also trackslist.h.                                                         *
 *                                                                            *
 * Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>                *
 *                                                                            *
 * This program is free software; you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published by the Free *
 * Software Foundation; either version 2 of the License, or (at your option)  *
 * any later version.                                                         *
 *                                                                            *
 * This program is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with this program. If not, see <http://www.gnu.org/licenses/>.             *
 ******************************************************************************/

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>

#include "compat.h"
#include "text.h"
#include "function.h"
#include "wordslist.h"
#include "trackslist.h"
#include "options.h"

using namespace std;


#define DEFRATE 2                         // default rating for unrated tracks
#define MAX_WORDS 0x3fff                  // maximum number of different words

extern Options options;

WordsList words = MAX_WORDS;              // list of words from track filenames


 TracksList::TracksList(int n) {
   /***************************************************************************
    * CONSTRUCTOR TracksList: allows for maximum of n tracks.                 *
    ***************************************************************************/
   maxTracks = n;
   nTracks = 0;
   totalSize = 0;

   /*
    * When track details are set (Set method), we use hashing to quickly see
    * if there is already a matching track in list[]. We need the size of
    * list[] to be a power of 2. We choose the next power of 2 that is larger
    * than maxTracks. This ensures there is at least one unused location in
    * list[] so we won't lock up when searching for a track that isn't there.
    */
   listSize = 1;
   while (listSize <= maxTracks)
      listSize = listSize + listSize;

   /*
    * We allocate list[] and initialise all hashes to zero to indicate unused
    * entries.
    */
   list = new trackslist[listSize];
   for (int i = 0; i < listSize; i++) {
      list[i].hash = 0; }

   /*
    * We allocate sequence. No initialisation is needed.
    */
   sequence = new int[maxTracks]; }



 TracksList::~TracksList() {
   /***************************************************************************
    * DESTRUCTOR TracksList                                                   *
    ***************************************************************************/

   /*
    * We find all used entries in list[] and deallocate their filePath
    * strings. Note that a hash of 1 indicates a deleted entry: its filePath
    * is already deallocated.
    */
   for (int i = 0; i < listSize; i++) {
      if ((unsigned) list[i].hash >= 2)
         delete[] list[i].filePath; }

   delete[] sequence;
   delete[] list; }



 void TracksList::Set(const char *filePath, int rating, const char *attrib,
   const char *exts, long long size) {
   /***************************************************************************
    * This checks if the track with path/filename filePath is already in      *
    * list[]. If it isn't, it adds it and sets its rating, attributes and     *
    * size. If it is, it changes its recorded size to size and, if it had a   *
    * lower rating, it updates its rating to rating. If no attributes were    *
    * set, it updates those too.                                              *
    ***************************************************************************/
   int hash,                                                // hash of filePath
      freepos = -1,                                  // free position in list[]
      pos,                                                // position in list[]
      ppos = 0;                                        // path parsing position

   char buf[20];

   if (!filePath || !*filePath)
      return;                                            // The path is invalid
   /*
    * We get and check the file extension.
    */
   PathParse(filePath, buf, sizeof buf, 0, &ppos);
   if (exts && !MatchExt(buf, exts))
      return;

   /*
    * We hash the track file path/name and use it to get an initial location
    * in list[].
    */
   pos = (hash = RSHash(filePath)) & (listSize - 1);

   /*
    * We ensure the hash is not zero or 1 (a zero hash in list[] indicates an
    * unused entry, 1 indicates a deleted entry). We then search for a
    * matching hash or an unused entry. Just in case two path/names produce
    * the same hash (c.f. birthday paradox), we compare strings too if the
    * hashes match.
    */
   if ((unsigned) hash < 2)
      hash = 2;
   while (list[pos].hash && (list[pos].hash != hash
         || strcmp(filePath, list[pos].filePath))) {
      if (list[pos].hash == 1)
         freepos = pos;                                // found a deleted entry
      if (++pos >= listSize)
         pos = 0; }

   if ((unsigned) list[pos].hash >= 2) {
      /*
       * The track was already in the list. We set its size, correct the
       * total size and add attributes if none existed before. If the track
       * was unrated or had a lower rating, we update its rating too.
       */
      totalSize = totalSize + size - list[pos].size;
      list[pos].size = size;
      if (!*list[pos].attrib)
         strlcpy(list[pos].attrib, attrib ?
            attrib : "", sizeof list[pos].attrib);
      if (list[pos].rating < rating)
         list[pos].rating = rating; }

   else if (nTracks < maxTracks) {
      /*
       * It isn't and there is room. We add it, allocating a new filePath.
       */
      if (freepos >= 0)
         pos = freepos;                              // replace a deleted entry
      int len = strlen(filePath) + 1;
      list[pos].filePath = new char[len];
      strlcpy(list[pos].filePath, filePath, len);
      list[pos].hash = hash;
      list[pos].size = size;
      list[pos].rating = rating;
      strlcpy(list[pos].attrib, attrib ?
         attrib : "", sizeof list[pos].attrib);
      list[pos].link = pos;
      totalSize += size;
      sequence[nTracks++] = pos;

      /*
       * We are already positioned at the file extension dot. We work
       * backwards, getting words from the file name. We will stop once we
       * reach a path separator.
       */
      int wpos = 0,                 // current position in words list for track
         wnum;                              // handle for the word in wordslist
      if (PathParse(filePath, buf, sizeof buf, -1, &ppos)) {
         do {
            /*
             * We get a handle for a word: this also increments the count for
             * the word.
             */
            if ((wnum = words.Set(buf)) && wpos < TRACK_WORDS)
               /*
                * The handle is valid and there is still room. We record the
                * word.
                */
               list[pos].words[wpos++] = wnum; }
         while (ppos && PathParse(filePath, buf, sizeof buf, 1, &ppos)); }

      /*
       * We mark remaining word positions as unused.
       */
      while (wpos < TRACK_WORDS)
         list[pos].words[wpos++] = 0; } }



 int TracksList::Purge(void) {
   /***************************************************************************
    * This removes any tracks which are unrated, have no attributes and are   *
    * zero-size (the file is missing or deleted). It returns the number of    *
    * tracks removed.                                                         *
    ***************************************************************************/
   int pos1 = 0,                                      // position in sequence[]
      pos2,                                               // position in list[]
      nRemoved = 0;                                 // number of tracks removed

   while (pos1 < nTracks) {
      pos2 = sequence[pos1];
      if (list[pos2].rating < 0 && !*list[pos2].attrib && !list[pos2].size) {
         Remove(pos1);
         nRemoved += 1; }
      else
         pos1 += 1; }

   return nRemoved; }



 void TracksList::UpdatePath(int pos, const char *newpath) {
   /***************************************************************************
    * This changes the file path associated with the track at pos to newpath. *
    *                                                                         *
    * IMPORTANT: hashes aren't updated. Don't use this function until all     *
    * tracks have been set.                                                   *
    ***************************************************************************/
   pos = sequence[pos];
   if ((unsigned) list[pos].hash >= 2) {
      delete[] list[pos].filePath; }
   int len = strlen(newpath) + 1;
   list[pos].filePath = new char[len];
   strlcpy(list[pos].filePath, newpath, len); }



 void TracksList::Shuffle(int n) {
   /***************************************************************************
    * This shuffles entries at positions 0 to n-1 in sequence[].              *
    ***************************************************************************/
   int mask,                                     // used to mask random numbers
      pos,                                    // current position in sequence[]
      swap;                                            // position to swap with

   pos = n - 1;
   while (pos > 0) {
      /*
       * We implement the Durstenfeld shuffle algorithm, working back from
       * the last entry to the second entry. First we generate a mask for the
       * random number.
       */
      mask = 1;
      while (mask < pos)
         mask = mask + mask + 1;

      /*
       * We get a random number and mask it. If it is too big, we try again.
       */
      do {
         swap = rand() & mask; }
      while (swap > pos);

      /*
       * We swap the current track with the randomly chosen one.
       */
      Swap(pos--, swap); } }



 int TracksList::UseTrack(int pos, int weighting) {
   /***************************************************************************
    * This decides whether to add the track at pos to the playlist. The       *
    * decision is based on its rating, weighting and a fresh random number.   *
    * It returns TRUE (nonzero) if the track is to be added.                  *
    ***************************************************************************/
   int approvalProb,              // 0xffff times probability of returning TRUE
      rating;                                               // the track rating

   const int weightFactor[] = {
      256, 208, 169, 137, 111,
      91, 74, 60, 49, 39, 32 };

   /*
    * We get the index for the track.
    */
   rating = list[pos = sequence[pos]].rating;

   /*
    * We make sure the weighting is in the valid range (0-10), then look up
    * the corresponding factor.
    */
   if (weighting < 0)
      weighting = 0;
   else if (weighting > 10)
      weighting = 10;
   weighting = weightFactor[weighting];

   approvalProb = 0xffff;                                                // 1.0
   /*
    * For each rating step below 6, we multiply the probability by the
    * multiplier and divide by 256. This effectively multiplies it by between
    * 1 and 1/8 (0.125) depending on the weighting.
    */
   while (rating++ < MAX_RATING)
      approvalProb = (weighting * approvalProb) >> 8;

   /*
    * We get a random number and mask it to the probability range, If it's
    * less than or equal to the probability, we indicate to use the track.
    */
   return (rand() & 0xffff) <= approvalProb; }



int TracksList::Match(int a, int b) {
   /***************************************************************************
    * This compares the file names for tracks a and b and returns a value     *
    * indicating how similar they are.                                        *
    ***************************************************************************/
   int pos_a = 0,                                  // word position for track a
      pos_b,                                       // word position for track b
      pos_p = TRACK_WORDS,                // previous matching track b word pos
      rslt = 0,                                                 // match result
      in_seq = 2,                           // 2 * number of sequential matches
      word;                                              // current word handle

   /*
    * We get the indexes for the two tracks.
    */
   a = sequence[a];
   b = sequence[b];

   while (pos_a < TRACK_WORDS && (word = list[a].words[pos_a])) {
      pos_a += 1;
      /*
       * For each word in track a, we search words in track b for a match.
       */
      pos_b = 0;
      while (pos_b < TRACK_WORDS && word != list[b].words[pos_b])
         pos_b += 1;

      if (pos_b < TRACK_WORDS) {
         /*
          * We found a matching word. We see if the previous words in the
          * track names matched too.
          */
         if (pos_b == pos_p + 1)
            in_seq += 2;                                                // Yes.
         else
            in_seq = 2;                                                  // No.
         pos_p = pos_b;

         /*
          * We add to the match index according to how many previous words
          * matched and how unusual this word is.
          */
         rslt += in_seq + (nTracks + 2048) / (words.Count(word) << 10); }

      else {
         /*
          * No matching word was found. We reset things.
          */
         in_seq = 2;
         pos_p = TRACK_WORDS; } }

   /*
    * We have finished matching words. We find how many words were in the
    * shorter of the two track names and subtract this from the match index
    * (penalising it if there were many possible matching words).
    */
   while (pos_a && !list[b].words[pos_a - 1])
      pos_a -= 1;
   rslt -= pos_a;

   return rslt; }



void TracksList::Link(int a, int b) {
   /***************************************************************************
    * This sets links for tracks a and b to link them in a list. If either or *
    * both are already in lists, the lists are merged.                        *
    ***************************************************************************/
   int firstTrack,                             // first track in resulting list
      x;                                                             // scratch

   /*
    * We find the first track in each list.
    */
   a = ListFindFirst(sequence[a]);
   b = ListFindFirst(sequence[b]);

   if (a == b)
      /*
       * The first tracks are the same: the lists are already linked.
       */
      return;

   /*
    * We find which will be the first track in the merged list, then find the
    * last track in each list.
    */
   firstTrack = (b > a) ? a : b;
   a = list[a].link;
   b = list[b].link;

   if (b > a) {
      /*
       * List b ends after list a. We swap them so that a will be the last
       * track in the merged list.
       */
      x = a;
      a = b;
      b = x; }

   /*
    * We set the link for the first track in the merged list to point to the
    * last one.
    */
   list[firstTrack].link = a;

   while (a > firstTrack) {
      /*
       * We haven't reached the first track in the merged list yet.
       */

      if (b < a && b > list[a].link) {
         /*
          * List b isn't finished and its last entry is after the previous
          * entry in list a. We get the previous entry in list a, link the
          * last entry in list a to the last entry in list b, then move to
          * the previous entry in list b.
          */
         x = list[a].link;
         list[a].link = b;

         if (b > x) {
            /*
             * The new last entry in list b is after the previous one in list
             * a. We swap the lists.
             */
            a = b;
            b = x; }

         else
            /*
             * The previous entry in list a is after the last one in list b.
             * We move to the previous entry in list a.
             */
            a = x; }
      else {
         /*
          * Either list b is finished or its last entry is before the
          * previous entry in list a.
          */
         if ((x = list[a].link) >= a)
            /*
             * List a is finished. We move to list b.
             */
            a = list[a].link = b;
         else
            /*
             * List a is not finished. We move to its previous entry.
             */
            a = x; } } }



void TracksList::Remove(int pos) {
   /***************************************************************************
    * This removes the track at pos from the list.                            *
    ***************************************************************************/
   int pos2;                                              // position in list[]

   if (list[pos2 = sequence[pos]].hash >= 2) {
      /*
       * The track exists. We set its hash to 1 (deleted entry) and delete
       * its file path. Note: we mustn't set the hash to zero (unused entry):
       * that could prevent subsequent hashes being matched.
       */
      list[pos2].hash = 1;
      delete[] list[pos2].filePath; }

   /*
    * We remove the track from the sequence.
    */
   nTracks -= 1;
   while (pos < nTracks) {
      sequence[pos] = sequence[pos + 1];
      pos += 1; } }



void TracksList::Swap(int a, int b) {
   /***************************************************************************
    * This swaps entries a and b in sequence[].                               *
    ***************************************************************************/
   int x = sequence[a];
   sequence[a] = sequence[b];
   sequence[b] = x; }



void TracksList::DefaultRate(int r) {
   /***************************************************************************
    * This finds unrated tracks and sets them to the default rating r.        *
    ***************************************************************************/
   int pos1 = 0,                                      // position in sequence[]
      pos2;                                               // position in list[]

   while (pos1 < nTracks) {
      if (list[pos2 = sequence[pos1++]].rating < 0)     // the track is unrated
         list[pos2].rating = r; } }



int TracksList::SortLinks() {
   /***************************************************************************
    * This sorts the tracks so that linked ones are together at the start of  *
    * the list. It returns the total number of tracks that are linked.        *
    ***************************************************************************/
   int dpos = 0,             // destination position in sequence[] when sorting
      cpos,                      // current position in sequence[] when sorting
      pos,                                                // position in list[]
      npos;                                 // next (linked) position in list[]

   do {
      /*
       * We work through sequence[]. All entries before dpos are already
       * sorted. We start at dpos and look for the first entry in a linked
       * list.
       */
      cpos = dpos;
      while (cpos < nTracks && (npos = list[pos = sequence[cpos]].link) <= pos)
         cpos += 1;

      if (cpos < nTracks) {
         /*
          * We found one. We move it.
          */
         Swap(dpos++, cpos);

         do {
            /*
             * npos is pointing to the next entry in the linked list. We look
             * for the corresponding entry in sequence[].
             *
             * We will repeat these steps until the next entry is the
             * original first entry.
             */
            cpos = dpos;
            while (cpos < nTracks && npos != sequence[cpos])
               cpos += 1;

            if (cpos < nTracks) {
               /*
                * We found the corresponding entry. We move it, then get the
                * next link.
                */
               Swap(dpos++, cpos);
               npos = list[npos].link; } }
         while (cpos < nTracks && npos != pos); } }

   while (cpos < nTracks);
   return dpos; }



int TracksList::ListFindFirst(int a) {
   /***************************************************************************
    * This private method finds the first entry in list[] that is linked to   *
    * entry a. Note that a and the return value are positions in list[]       *
    * itself in contrast to public methods where positions are translated via *
    * sequence[].                                                             *
    ***************************************************************************/
   while (list[a].link < a)
      a = list[a].link;
   return a; }




istream &operator>>(istream &stream, TracksList &tl) {
   /***************************************************************************
    * EXTRACTOR TracksList: This reads in all of tl in CSV format from        *
    * stream.                                                                 *
    ***************************************************************************/
   char buf[MAX_FP_CHARS+200],                         // holds entire CSV line
      buf2[MAX_FP_CHARS];                          // holds unescaped file path

   while (!stream.eof()) {
      stream.getline(buf, sizeof buf);
      CSVLine* csv = new CSVLine(buf);
      tl.Set(UnEscapePath(buf2, csv->GetFieldAsString(0, NULL)),
         csv->GetFieldAsInt(1, -1), csv->GetFieldAsString(2, ""), NULL, 0);
      delete csv; }

   return stream; }



std::ostream &operator<<(std::ostream &stream, TracksList &tl) {
   /***************************************************************************
    * INSERTER TracksList: This outputs all of tl in CSV format to stream,    *
    * omitting unrated tracks where the size is zero.                         *
    ***************************************************************************/
   int pos,                                           // position in sequence[]
      pos2,                                               // position in list[]
      npos,                                 // next (linked) position in list[]
      matchNum = 0;                                   //sequential match number

   char buf[MAX_FP_CHARS];                           // holds escaped file path

   for (pos = 0; pos < tl.GetNumber(); pos++) {
      /*
       * We work through sequence[]. We check the link for the current track.
       */
      pos2 = tl.sequence[pos];
      if ((npos = tl.list[pos2].link) != pos2) {
         /*
          * The track isn't just linked to itself: it is part of a linked
          * list. If its link points forward, it is the first one in the list
          * and we increment the match number.
          */
         if (npos > pos2)
            matchNum += 1; }

      else
         /*
          * The track is linked to itself: it isn't part of a linked list. We
          * should have used SortLinks() to move all lists to the start so
          * they must all be done now. We indicate not to show a match
          * number.
          */
         matchNum = 0;

      stream << "\"" << EscapePath(buf, tl.list[pos2].filePath,
         sizeof buf, options.excelMode) << "\",";
      if (tl.list[pos2].rating >= 0)
         stream << tl.list[pos2].rating;
      stream << ",";
      if (*tl.list[pos2].attrib)
         stream << tl.list[pos2].attrib;
      stream << ",";
      if (tl.list[pos2].size > 0)
         stream << tl.list[pos2].size;
      if (matchNum)
         stream << "," << matchNum;
      stream << endl; }

   return stream; }

