/******************************************************************************
 * trackslist.h - Headers for class that maintains a list of tracks. See      *
 * trackslist.cpp.                                                            *
 *                                                                            *
 * Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>                *
 *                                                                            *
 * This program is free software; you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published by the Free *
 * Software Foundation; either version 2 of the License, or (at your option)  *
 * any later version.                                                         *
 *
 * This program is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with this program. If not, see <http://www.gnu.org/licenses/>.             *
 ******************************************************************************/

#define TRACK_WORDS 10
#define MAX_WORDLEN 15

class TracksList {
   /***************************************************************************
    * This implements an ordered list of tracks. A maximum number of tracks   *
    * to accommodate is specified when it is constructed. The Set method adds *
    * a track or updates its details. There are methods to do various things  *
    * including getting details of an individual track or information about   *
    * the total list.                                                         *
    *                                                                         *
    * A track has a set position in list[] based on the hash value of the     *
    * file path and name. The list ordering is determined by sequence[] which *
    * translates list position to a position in list[]. E.g. the first and    *
    * second tracks in the ordered list can be swapped by swapping the values *
    * (pointing into list[]) stored in sequence[0] and sequence[1].           *
    ***************************************************************************/

   struct trackslist {
      char *filePath,   // path (relative to music directory) to the music file
         attrib[11];                                             // disposition
      long long size;                        // size of the music file in bytes
      int hash,                                         // the hash of filePath
         rating,            // -1 (unrated), 0 (don't play) or 1-5 (worst-best)
         link,                      // to previous or last track record in list
         words[TRACK_WORDS]; }                     // words found in track name
   *list;                                         // hash-ordered track details

   long long
      totalSize;                    // total size of all the tracks in the list

   int
      listSize,                // size of list[] (may be larger than maxTracks)
      maxTracks,                            // maximum number of tracks allowed
      nTracks,                                      // current number of tracks
      *sequence;                               // pointers to entries in list[]


   public:

   /*
    * The INSERTER (in trackslist.cpp - outputs to a file) needs access tp
    * private members. Note that the EXTRACTOR doesn't: it just uses the
    * constructor.
    */
   friend std::ostream &operator<<(std::ostream &stream, TracksList &tl);

   TracksList(int n);
   ~TracksList();

   void Set(const char *filePath, int rating, const char *attrib,
      const char *exts, long long size);                 // checks/adds a track
   void Shuffle(int n);                    // random shuffle of tracks 0 to n-1
   int Purge(void);                            // purges unneeded track entries
   void UpdatePath(int pos, const char *newpath);        // changes a path name
   int UseTrack(int pos, int weighting);              // probabilistic decision
   int Match(int a, int b);     // matching index between file names of a and b
   void Link(int a, int b);                   // links tracks a and b in a list
   void DefaultRate(int r);               // assigns rating r to unrated tracks
   void Remove(int pos);                                // removes track at pos
   void Swap(int a, int b);                             // swaps tracks a and b
   int SortLinks();            // puts tracks linked in lists together at start



   void Shuffle() {
      /************************************************************************
       * This shuffles all the tracks.                                        *
       ************************************************************************/
      Shuffle(nTracks); }


   /***************************************************************************
    * The following public methods return information about track n (starting *
    * with 0).                                                                *
    ***************************************************************************/
   char *GetPath(int n) {                             // its file path and name
      return list[sequence[n]].filePath; }

   int GetRating(int n) {                                // its assigned rating
      return list[sequence[n]].rating; }

   const char *GetAttrs(int n) {                               // its attribute
      return list[sequence[n]].attrib; }

   long long GetSize(int n) {                         // its file size in bytes
      return list[sequence[n]].size; }

   int FirstLink(int n) {    // if it is the first or only track in linked list
      n = sequence[n];
      return n <= list[n].link; }


   /***************************************************************************
    * The following public methods return information about the overall list. *
    ***************************************************************************/
   int GetNumber() {                            // number of tracks in the list
      return nTracks; }

   long long GetTotalSize() {              // total size of all the track files
      return totalSize; }

   private:

   int ListFindFirst(int a);          // finds the first track in a linked list

 };                                                           // END TracksList



std::istream &operator>>(std::istream &stream, TracksList &tl);    // EXTRACTOR
std::ostream &operator<<(std::ostream &stream, TracksList &tl);     // INSERTER
