/******************************************************************************
 * wordslist.cpp - Class that maintains a list of words with a count for      *
 * each. See also wordslist.h.                                                *
 *                                                                            *
 * Copyright (C) Steve Hayes 2011-2013 <steve@honeylink.co.uk>                *
 *                                                                            *
 * This program is free software; you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published by the Free *
 * Software Foundation; either version 2 of the License, or (at your option)  *
 * any later version.                                                         *
 *                                                                            *
 * This program is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with this program. If not, see <http://www.gnu.org/licenses/>.             *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <string.h>

#include <iostream>

#include "compat.h"
#include "wordslist.h"
#include "function.h"



WordsList::WordsList(int n) {
   /***************************************************************************
    * CONSTRUCTOR WordsList: allows for maximum of n words.                   *
    ***************************************************************************/
   maxWords = n;
   nWords = 0;

   /*
    * When a word is set (Set method), we use hashing to quickly see if there
    * is already a matching word in list[]. We need the size of list[] to be
    * a power of 2. We choose the next power of 2 that is larger than
    * maxWords. (This ensures there is at least one unused location in list[]
    * so we won't lock up when searching for a word that isn't there.)
    */
   listSize = 1;
   while (listSize <= maxWords)
      listSize = listSize + listSize;

   /*
    * We allocate list[] and initialise all hashes to zero to indicate unused
    * entries.
    */
   list = new wordslist[listSize];
   for (int pos = 0; pos < listSize; pos++) {
      list[pos].hash = 0; } }



 WordsList::~WordsList() {
   /***************************************************************************
    * DESTRUCTOR WordsList                                                    *
    ***************************************************************************/
   delete[] list; }



 int WordsList::Set(const char *word) {
   /***************************************************************************
    * This checks if word is already in list[]. If it isn't, it adds it and   *
    * sets its count to 1. If it is, it increments its count. In either case, *
    * it returns its position in list[] plus 1. (This can be used later with  *
    * Count to find the total incidences of the word.)                        *
    *                                                                         *
    * If word is too short or if list[] is full, list[] and the counts are    *
    * unchanged and 0 is returned.                                            *
    ***************************************************************************/
   char cchar,                                             // current character
      buf[100];                                 // word converted to upper case
   int pos,                                       // position in buf, then list
      hash;                                          // hash of upper-case word

   if (strlen(word) < 3)
      return 0;                                            // word is too short

   /*
    * We copy word to buf, making it upper case (to make things
    * case-insensitive).
    */
   pos = 0;
   do {
      cchar = word[pos];
      if (cchar >= 'a' && cchar <= 'z')
         cchar -= 'a' - 'A';
      if (pos < sizeof buf)
         buf[pos] = cchar;
      pos += 1; }
   while (cchar);

   /*
    * We ensure the word is null-terminated even if it is too long.
    */
   buf[sizeof buf - 1] = '\0';

   /*
    * We hash the word and use it to get an initial location in list[].
    */
   pos = (hash = RSHash(buf)) & (listSize - 1);

   /*
    * We ensure the hash is non-zero (a zero hash in list[] indicates an
    * unused entry). We then search for a matching hash or an unused entry.
    */
   if (!hash)
      hash = 1;
   while (list[pos].hash && list[pos].hash != hash) {
      if (++pos >= listSize)
         pos = 0; }

   if (list[pos].hash)
      /*
       * The word was already in the list. We increment its count.
       */
      list[pos].count += 1;

   else if (nWords < maxWords) {
      /*
       * It isn't and there is room. We add it.
       */
      list[pos].hash = hash;
      list[pos].count = 1;
      nWords += 1; }

   else
      /*
       * The list is full.
       */
      return 0;

   return pos + 1; }
